library fluttershy_backend_dartex;

import 'package:dartex/dartex.dart';
import 'package:flutter/widgets.dart' hide Transform, Size;
import 'package:fluttershy/backend.dart';
import 'package:fluttershy/event.dart';
import 'package:fluttershy/events/resize_event.dart';
import 'package:fluttershy_backend_dartex/components/children.dart';
import 'package:fluttershy_backend_dartex/components/local_to_world.dart';
import 'package:fluttershy_backend_dartex/components/parent.dart';
import 'package:fluttershy_backend_dartex/components/transform.dart';
import 'package:fluttershy_backend_dartex/resources/renderer.dart';
import 'package:fluttershy_backend_dartex/resources/screen_dimensions.dart';
import 'package:fluttershy_backend_dartex/systems/camera_system.dart';
import 'package:fluttershy_backend_dartex/systems/rectangle_rendering_system.dart';
import 'package:fluttershy_backend_dartex/systems/transform_system.dart';

class DartexBackend extends Backend {
  final Function(BuildContext context, World world) init;
  final List<System> systems;
  final World _world;
  final bool _setupDefaults;

  DartexBackend({
    List<System> systems,
    @required this.init,
    bool setupDefaults = true,
  })  : systems = systems ?? [],
        _world = World(
          systems: setupDefaults
              ? [
                  ...systems,
                  TransformSystem(),
                  CameraSystem(),
                  RectangleRenderingSystem(),
                ]
              : systems,
        ),
        _setupDefaults = setupDefaults;

  @override
  void setup(BuildContext context) {
    if (_setupDefaults) {
      _world.addEventListener(_worldEventListener);
      _world.addResource(Renderer());
      _world.addResource(ScreenDimensions());
    }

    init(context, _world);
  }

  @override
  void event(Event event) {
    if (event is ResizeEvent) {
      _world.getResource<ScreenDimensions>().size = event.size;
    }
  }

  @override
  void update(double deltaTime) {
    _world.run();
  }

  @override
  void render(Canvas canvas) {
    _world.getResource<Renderer>().render(canvas);
  }

  @override
  void destroy(BuildContext context) {
    if (_setupDefaults) {
      _world.removeResource<ScreenDimensions>();
      _world.removeResource<Renderer>();
      _world.removeEventListener(_worldEventListener);
    }

    _world.destroyEntities();
    _world.destroyResources();
  }

  void _worldEventListener(WorldEvent event) {
    final entity = event.entity;

    if (event is ComponentAddedEvent) {
      switch (event.componentType) {
        case Transform:
          entity.addComponent(LocalToWorld());
          break;
        case Parent:
          final parent = entity.getComponent<Parent>();
          final children = parent.entity.getComponent<Children>();

          if (children != null) {
            children.entities.add(entity);
          } else {
            parent.entity.addComponent(Children([entity]));
          }
          break;
      }
    } else if (event is ComponentReplacedEvent) {
      switch (event.componentType) {
        case Parent:
          final oldParent = event.oldComponent as Parent;
          final newParent = entity.getComponent<Parent>();

          oldParent.entity
              .getComponent<Children>()
              .entities
              .removeWhere((child) => child.id == entity.id);

          final children = newParent.entity.getComponent<Children>();

          if (children != null) {
            children.entities.add(entity);
          } else {
            newParent.entity.addComponent(Children([entity]));
          }

          break;
      }
    } else if (event is ComponentRemovedEvent) {
      switch (event.componentType) {
        case Transform:
          entity.removeComponent<LocalToWorld>();
          break;
        case Parent:
          final parent = event.removedComponent as Parent;
          final children = parent.entity.getComponent<Children>();

          children.entities.removeWhere((child) => child.id == entity.id);

          if (children.entities.isEmpty) {
            parent.entity.removeComponent<Children>();
          }

          break;
      }
    }
  }
}
